
import os
import sys
from .functions import definitions
from .expression import add_context_processor
from .builder import build_all
from .dumper import dump


for definition in definitions:
    add_context_processor(definition)


def process(input, output, filename=None, errors=[]):
    if filename:
        filename = os.path.abspath(filename)    
    first = True

    for doc in build_all(input, filename, errors):
        if not first:
            output.write("---\n")
        else:
            first = False
        dump(doc, output, errors)




''' Transforms the parsed YAML structure into a tree '''

from . import loader
from .tree import (Document, Scalar, Sequence, Mapping, MergeKey)
from .expression import (Expression, Function, FunctionBlock)
from ruamel.yaml.nodes import (ScalarNode, SequenceNode, MappingNode)
from ruamel.yaml.comments import TaggedScalar


def build(source, filepath=None):
    yaml_doc = loader.load(source)
    return create_tree(yaml_doc, filepath)


def build_all(source, filepath=None, errors=[]):
    for yaml_doc in loader.load_all(source, errors):
        yield create_tree(yaml_doc, filepath)


def create_tree(yaml_doc, filepath):
    if yaml_doc is None:
        return None
    doc = Document()
    tree = convert(yaml_doc, doc)
    doc.filepath = filepath
    doc.root = tree
    return tree


def convert(node, doc):
    if isinstance(node, dict):
        return Mapping([(convert(k, doc), convert(v, doc)) for (k, v) in node.items()], doc)
    if isinstance(node, list):
        return Sequence([convert(v, doc) for v in node], doc)
    if isinstance(node, TaggedNode):
        return node.create(doc)
    return Scalar(node, doc)


def construct_node(self, node):
    if isinstance(node, ScalarNode):
        return self.construct_scalar(node)
    if isinstance(node, MappingNode):
        return [c for c in self.construct_yaml_map(node)][0]
    if isinstance(node, SequenceNode):
        result = [c for c in self.construct_yaml_seq(node)][0]
        return result


class TaggedNode:
    def __init__(self, value):
        self.value = value


class InvisibleNode(TaggedNode):
    def create(self, doc):
        if isinstance(self.value, str):
            result = Expression(self.value, doc)
        else:
            result = convert(self.value, doc)
        result.visible = False
        return result


def invisible_constructor(self, node):
    return InvisibleNode(construct_node(self, node))


loader.add_constructor('!-', invisible_constructor)


class ExpressionNode(TaggedNode):
    def create(self, doc):
        result = Expression(self.value, doc)
        return result


def expression_constructor(self, node):
    if isinstance(node, ScalarNode):
        return ExpressionNode(self.construct_scalar(node))
    assert False, 'expression nodes must be scalar'


loader.add_constructor('!?', expression_constructor)


class FunctionNode(TaggedNode):
    def create(self, doc):
        result = Function(self.value, doc)
        return result


class FunctionBlockNode(TaggedNode):
    def __init__(self, value, constructor):
        self.value = value
        self.constructor = constructor
        self.node = construct_node(self.constructor, self.value)

    def create(self, doc):
        result = FunctionBlock(lambda: convert(self.node, doc))
        return result


def function_constructor(self, node):
    if isinstance(node, ScalarNode):
        return FunctionNode(self.construct_scalar(node))
    else:
        return FunctionBlockNode(node, self)


loader.add_constructor('!()', function_constructor)


class IfNode(TaggedNode, MergeKey):
    def __init__(self, doc):
        TaggedNode.__init__(self, doc)
        MergeKey.__init__(self)

    def create(self, doc):
        self.expression = Expression(self.value, doc)
        return self

    def merge(self, parent, source):
        if self.expression.content():
            MergeKey.merge(self, parent, source)


def if_constructor(self, node):
    if isinstance(node, ScalarNode):
        return IfNode(self.construct_scalar(node))
    assert False, 'expression nodes must be scalar'


loader.add_constructor('!if', if_constructor)


class ForNode(TaggedNode, MergeKey):
    def __init__(self, doc):
        TaggedNode.__init__(self, doc)
        MergeKey.__init__(self)

    def create(self, doc):
        self.expression = Expression(self.value, doc)
        return self

    def merge(self, parent, source):
        for item in self.expression.content():
            MergeKey.merge(self, parent, source.content()(item))


def for_constructor(self, node):
    if isinstance(node, ScalarNode):
        return ForNode(self.construct_scalar(node))
    assert False, 'expression nodes must be scalar'


loader.add_constructor('!for', for_constructor)



''' Command-line interface '''

import sys
from . import version
from .main import process

class ErrorList(list):
    def append(self, error):
        super().append(error)
        sys.stderr.write("---\n")
        sys.stderr.write(str(error)+"\n")
        sys.stderr.flush()

def main():
    errors = ErrorList()
    if len(sys.argv) > 1:
        if sys.argv[1] == '-v':
            print('yglu '+version)
            sys.exit(0)
        if sys.argv[1].startswith('-'):
            print('Usage: yglu [options] [<filename>]')
            print()
            print('Options:')
            print('  -v - -version          Print version and exit.')
            print('  -h - -help             Print help and exit.')
            sys.exit(0)
        try:
            filename = sys.argv[1]
            with open(filename) as file:
                process(file, sys.stdout, filename, errors)
        except FileNotFoundError:
            sys.stderr.write('File not found: '+sys.argv[1]+'\n')
            sys.exit(1)
    else:
        process(sys.stdin, sys.stdout, None, errors)
    if len(errors) > 0:
        sys.stderr.write("Finished with errors.\n")
        sys.exit(1)



''' Model for the internal document structure. '''

from collections import OrderedDict


class Document:
    def __init__(self):
        self.filepath = None
        self.root = None
        self.currentPath = []


class Node:
    def __init__(self, doc):
        self.visible = True
        self.memo = None
        self.doc = doc

    def content(self):
        if(self.memo is None):
            self.memo = self.create_content()
        return self.memo

    def create_content(self):
        return self


class MergeKey:
    def __init__(self):
        self.visited = False

    def merge(self, parent, source):
        self.visited = True
        if isinstance(parent, Mapping):
            if isinstance(source, OrderedDict):
                parent.update(OrderedDict.items(source))
            elif isinstance(source, dict):
                parent.update(source)               
            if isinstance(source, Mapping):
                parent.special_entries.extend(source.special_entries)
        else:
            parent.extend(source)

class Scalar(Node):
    def __init__(self, value, doc=None):
        Node.__init__(self, doc)
        self.memo = value

    def content(self):
        return self.memo

    def __repr__(self):
        return str(self.memo)


class Mapping(OrderedDict, Node):
    def __init__(self):
        Node.__init__(self, None)

    def __init__(self, source=None, doc=None):
        Node.__init__(self, doc)
        self.special_entries = []
        if source:
            if isinstance(source, OrderedDict):
                source = OrderedDict.items(source)
            elif isinstance(source, dict):
                source = dict.items(source)
            OrderedDict.__init__(self, self.handle_keys(source))


    def __getitem__(self, key):
        if(self.__contains__(key)):
            return super().__getitem__(key).content()
        else:
            return self.resolve_special(key)

    def items(self):
        self.resolve_special()
        items = filter(lambda kv: kv[1].visible and kv[1].content() is not None,
                       super().items())
        return [(k, v.content()) for (k, v) in items]

    def __eq__(self, other):
        return dict(self.items()) == other

    def __repr__(self):
        return 'y'+dict(self.items()).__repr__()

    def handle_keys(self, items):
        for (k, v) in items:
            if isinstance(k, Scalar):
                yield (k.content(), v)
            else:
                self.special_entries.insert(0, (k, v))

    def resolve_special(self, key=None):
        while len(self.special_entries) > 0:
            (k, v) = self.special_entries.pop()
            if isinstance(k, MergeKey):
                k.merge(self, v)
            else:
                computed_key = k.content()
                if computed_key is not None:
                    self[computed_key] = v
                else:
                    raise Exception("mapping key is null")
        self.special_entries = []
        if key is not None:
            return super().__getitem__(key)


class Sequence(list, Node):
    def __init__(self, value=None, doc=None):
        Node.__init__(self, doc)
        if value:
            list.__init__(self, value)

    def __getitem__(self, index):
        return super().__getitem__(index).content()

    def __eq__(self, other):
        return list(self) == other

    def __iter__(self):
        iter = super().__iter__()
        try:
            while True:
                node = iter.__next__()
                if isinstance(node, Mapping) and len(node) == 0 and len(node.special_entries) == 1:
                    (k,v) = node.special_entries[0]
                    if not k.visited:
                        k.merge(self, v)
                elif isinstance(node, Node):
                    if node.visible:
                        yield node.content()
                else:
                    yield node
        except StopIteration:
            return

    def __repr__(self):
        return 'y'+list(self).__repr__()



''' Evaluation of YAQL expressions '''

import os
import yaql
from ruamel.yaml.nodes import (MappingNode, SequenceNode)
from .tree import Node

engine = yaql.factory.YaqlFactory(allow_delegates=True).create(options={
    'yaql.convertInputData': False,
})
scopes = []
stack = []
contexts = {}
context_processors = []

def add_context_processor(processor):
    context_processors.append(processor)



def get_context(root):
    key = id(root)
    if key in contexts:
        context = contexts[key]
    else:
        context = yaql.create_context(delegates=True)
        if 'YGLU_ENABLE_ENV' in os.environ:
            context['$env'] = os.environ
        context['$_'] = root
        for process in context_processors:
            process(context, root)
        contexts[key] = context
    return context


def push_stack(node):
    if node in stack:
        raise CircularReferenceException()
    stack.append(node)


def pop_stack():
    stack.pop()


def push_scope(scope):
    scopes.append(scope)


def pop_scope():
    scopes.pop()

class Holder:
    def __init__(self, value):
        self.value = value

class Expression(Node):
    def __init__(self, expression, doc):             
        Node.__init__(self, doc)
        if expression.startswith('.'):
            self.expression = '$_'+expression
        else:
            self.expression = expression
        self.doc = doc

    def create_content(self):
        push_stack(self)
        context = get_context(self.doc.root).create_child_context()
        if len(scopes) > 0:
            context['$'] = scopes[-1]        
        result = engine(self.expression).evaluate(context=context)
        pop_stack()
        if isinstance(result, Holder):
            return result.value
        else:
            return result


class Function(Node):
    def __init__(self, expression, doc):
        Node.__init__(self, doc)
        self.expression = expression
        self.visible = False

    def eval(self, scope=None):
        context = get_context(self.doc.root).create_child_context()
        context['$'] = scope
        result = engine(self.expression).evaluate(context=context)
        return result

    def create_content(self):
        return self.eval


class FunctionBlock(Node):
    def __init__(self, constructor):
        Node.__init__(self, None)
        self.visible = False
        self.constructor = constructor

    def eval(self, scope):
        push_scope(scope)        
        node = self.constructor()
        if isinstance(node, dict):
            result = dict(node.items())
        else:
            result = list(node)
        str(result) # force content creation
        pop_scope()
        return result

    def create_content(self):
        return self.eval


class CircularReferenceException(Exception):
    def __init__(self):
        super().__init__("circular reference in expression")



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("yglu/__init__.py", "r") as fh:
    exec(fh.read())

setup(
    name='yglu',
    version=version,
    description='YAML glue for structural templating and processing',
    author='Laurent Bovet',
    author_email='laurent.bovet@windmaster.ch',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/lbovet/yglu",
    packages=["yglu"],
    entry_points={
        "console_scripts": ["yglu=yglu.cli:main"]
    },
    install_requires=[
        "ruamel.yaml==0.16.5",
        "yaql==1.1.3"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



''' Serialization into YAML text '''

import ruamel.yaml
from .loader import SimpleString
from .tree import (Mapping, Sequence, Scalar)

yaml = ruamel.yaml.YAML()


def represent_mapping(representer, mapping):
    return representer.represent_mapping(u'tag:yaml.org,2002:map', mapping)


def represent_sequence(representer, sequence):
    return representer.represent_sequence(u'tag:yaml.org,2002:seq', sequence)


def represent_string(representer, string):
    return representer.represent_scalar(u'tag:yaml.org,2002:str', string)


yaml.representer.add_representer(Mapping, represent_mapping)
yaml.representer.add_representer(Sequence, represent_sequence)
yaml.representer.add_representer(SimpleString, represent_string)


def dump(tree, output, errors=[]):    
    if tree is not None:
        try:
            yaml.dump(tree, output)
        except Exception as error:
            errors.append(error)



''' Built-in functions  '''

import os
from .builder import build
from .expression import Holder

def import_definition(context, root):
    if hasattr(root, 'doc') and  hasattr(root.doc, 'filepath') and root.doc.filepath:
        dir = os.path.dirname(root.doc.filepath)
    else:
        dir = os.getcwd()

    def import_function(filename):
        if os.path.isabs(filename):
            filepath = filename
        else:
            filepath = os.path.join(dir, filename)
        with open(filepath) as file:
            result = build(file, filename)
            return Holder(result)

    context['$import'] = import_function

definitions = [import_definition]



''' Loads YAML structure from text input '''

import sys
import io
import ruamel.yaml


class String:
    pass


class SimpleString(ruamel.yaml.scalarstring.ScalarString, String):
    __slots__ = ('lc')

    style = ""

    def __new__(cls, value):
        return ruamel.yaml.scalarstring.ScalarString.__new__(cls, value)


class PreservedScalarString(ruamel.yaml.scalarstring.PreservedScalarString, String):
    __slots__ = ('lc')


class DoubleQuotedScalarString(ruamel.yaml.scalarstring.DoubleQuotedScalarString, String):
    __slots__ = ('lc')


class SingleQuotedScalarString(ruamel.yaml.scalarstring.SingleQuotedScalarString, String):
    __slots__ = ('lc')


class MyConstructor(ruamel.yaml.constructor.RoundTripConstructor):
    def construct_scalar(self, node):
        if not isinstance(node, ruamel.yaml.nodes.ScalarNode):
            raise ruamel.yaml.constructor.ConstructorError(
                None, None,
                'expected a scalar node, but found %s' % node.id,
                node.start_mark)

        if node.style == '|' and isinstance(node.value, ruamel.yaml.compat.text_type):
            ret_val = PreservedScalarString(node.value)
        elif bool(self._preserve_quotes) and isinstance(node.value, ruamel.yaml.compat.text_type):
            if node.style == "'":
                ret_val = SingleQuotedScalarString(node.value)
            elif node.style == '"':
                ret_val = DoubleQuotedScalarString(node.value)
            else:
                ret_val = SimpleString(node.value)
        else:
            ret_val = SimpleString(node.value)
        ret_val.lc = ruamel.yaml.comments.LineCol()
        ret_val.lc.line = node.start_mark.line
        ret_val.lc.col = node.start_mark.column
        return ret_val

def create_loader():
    yaml = ruamel.yaml.YAML()
    yaml.Constructor = MyConstructor
    yaml.allow_duplicate_keys = True
    return yaml


def add_constructor(tag, constructor):
    MyConstructor.add_constructor(tag, constructor)


def load_all(source, errors=[]):
    if source == sys.stdin:
        buffer = io.StringIO()
        for line in source:
            if line.replace(' ','').startswith('---'):
                yield create_loader().load(buffer.getvalue())
                buffer = io.StringIO()
            else:
                buffer.writelines(line)
        
        try:
            yield create_loader().load(buffer.getvalue())
        except Exception as error:
            errors.append(error)
    else:
        try:
            for doc in create_loader().load_all(source):
                yield doc
        except Exception as error:
            errors.append(error)

def load(source):
    return create_loader().load(source)
